library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package axi is
    subtype axi_id     is std_logic_vector(5 downto 0);
    subtype axi_addr   is std_logic_vector(31 downto 0);
    subtype axi_data   is std_logic_vector(63 downto 0);
    subtype axi_strobe is std_logic_vector(7 downto 0);
    subtype axi_len    is natural range 1 to 16;
    subtype axi_size   is std_logic_vector(1 downto 0);

    type axi_burst is ( axi_burst_fixed
                      , axi_burst_incr
                      , axi_burst_wrap
                  );
    
    type axi_resp is ( axi_resp_okay
                     , axi_resp_exokay
                     , axi_resp_slverr
                     , axi_resp_decerr
                 );

    type axi_lock is ( axi_lock_normal
                     , axi_lock_exclusive
                     , axi_lock_locked
                 );

    type axi_cache is
        record
            bufferable  : std_logic;
            cachable    : std_logic;
            read_alloc  : std_logic;
            write_alloc : std_logic;
        end record;

    type axi_prot is
        record
            priviliged  : std_logic;
            non_secure  : std_logic;
            instruction : std_logic;
        end record;

    function axi_burst_to_vector (burst : axi_burst) return std_logic_vector;
    function axi_resp_to_vector (resp : axi_resp) return std_logic_vector;
    function axi_len_to_vector (len : axi_len) return std_logic_vector;
    function axi_lock_to_vector (lock : axi_lock) return std_logic_vector;
    function axi_cache_to_vector (cache : axi_cache) return std_logic_vector;
    function axi_prot_to_vector (prot : axi_prot) return std_logic_vector;

    function vector_to_axi_burst (vector : std_logic_vector(1 downto 0)) return axi_burst;
    function vector_to_axi_resp  (vector : std_logic_vector(1 downto 0)) return axi_resp;
    function vector_to_axi_len   (vector : std_logic_vector) return axi_len;
    function vector_to_axi_lock  (vector : std_logic_vector(1 downto 0)) return axi_lock;
    function vector_to_axi_cache (vector : std_logic_vector(3 downto 0)) return axi_cache;
    function vector_to_axi_prot  (vector : std_logic_vector(2 downto 0)) return axi_prot;

    type axi_addr_channel is
        record
            id    : axi_id;
            addr  : axi_addr;
            len   : axi_len;
            size  : axi_size;
            burst : axi_burst;
            lock  : axi_lock;
            cache : axi_cache;
            prot  : axi_prot;
            valid : std_logic;
        end record;

    type axi_write_resp_channel is
        record
            id    : axi_id;
            resp  : axi_resp;
            valid : std_logic;
        end record;

    type axi_write_data_channel is
        record
            id    : axi_id;
            data  : axi_data;
            strb  : axi_strobe;
            last  : std_logic;
            valid : std_logic;
        end record;

    type axi_read_data_channel is
        record
            id    : axi_id;
            data  : axi_data;
            resp  : axi_resp;
            last  : std_logic;
            valid : std_logic;
        end record;
end axi;

package body axi is
    function axi_resp_to_vector (resp : axi_resp) return std_logic_vector is
    begin
        case resp is
            when axi_resp_okay   => return b"00";
            when axi_resp_exokay => return b"01";
            when axi_resp_slverr => return b"10";
            when axi_resp_decerr => return b"11";
        end case;
    end function;

    function axi_burst_to_vector (burst : axi_burst) return std_logic_vector is
    begin
        case burst is
            when axi_burst_fixed => return b"00";
            when axi_burst_incr  => return b"01";
            when axi_burst_wrap  => return b"10";
        end case;
    end function;

    function axi_len_to_vector (len : axi_len) return std_logic_vector is
    begin
        return std_logic_vector(to_unsigned(len - 1, 4));
    end function;

    function axi_lock_to_vector (lock : axi_lock) return std_logic_vector is
    begin
        case lock is
            when axi_lock_normal    => return b"00";
            when axi_lock_exclusive => return b"01";
            when axi_lock_locked    => return b"10";
        end case;
    end function;

    function axi_cache_to_vector (cache : axi_cache) return std_logic_vector is
        variable ret : std_logic_vector(3 downto 0);
    begin
        ret(0) := cache.bufferable;
        ret(1) := cache.cachable;
        ret(2) := cache.read_alloc;
        ret(3) := cache.write_alloc;
        return ret;
    end function;

    function axi_prot_to_vector (prot : axi_prot) return std_logic_vector is
        variable ret : std_logic_vector(3 downto 0);
    begin
        ret(0) := prot.priviliged;
        ret(1) := prot.non_secure;
        ret(2) := prot.instruction;
        return ret;
    end function;

    function vector_to_axi_resp (vector : std_logic_vector(1 downto 0)) return axi_resp is
    begin
        case vector is
            when b"00"  => return axi_resp_okay;
            when b"01"  => return axi_resp_exokay;
            when b"10"  => return axi_resp_slverr;
            when b"11"  => return axi_resp_decerr;
            when others => return axi_resp_okay;
        end case;
    end function;

    function vector_to_axi_burst (vector : std_logic_vector(1 downto 0)) return axi_burst is
    begin
        case vector is
            when b"00"  => return axi_burst_fixed;
            when b"01"  => return axi_burst_incr;
            when b"10"  => return axi_burst_wrap;
            when others => return axi_burst_fixed;
        end case;
    end function;

    function vector_to_axi_len (vector : std_logic_vector) return axi_len is
    begin
        return to_integer(unsigned(vector)) + 1;
    end function;

    function vector_to_axi_lock  (vector : std_logic_vector(1 downto 0)) return axi_lock is
    begin
        case vector is
            when b"00"  => return axi_lock_normal;
            when b"01"  => return axi_lock_exclusive;
            when b"10"  => return axi_lock_locked;
            when others => return axi_lock_normal;
        end case;
    end function;

    function vector_to_axi_cache (vector : std_logic_vector(3 downto 0)) return axi_cache is
        variable ret : axi_cache;
    begin
        ret.bufferable  := vector(0);
        ret.cachable    := vector(1);
        ret.read_alloc  := vector(2);
        ret.write_alloc := vector(3);
        return ret;
    end function;

    function vector_to_axi_prot  (vector : std_logic_vector(2 downto 0)) return axi_prot is
        variable ret : axi_prot;
    begin
        ret.priviliged  := vector(0);
        ret.non_secure  := vector(1);
        ret.instruction := vector(2);
        return ret;
    end function;
end package body;
