-----------------------------------------------------------------------------
-- AXI Demultiplexer
-----------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use work.axi.all;


entity axi_demux is
    port (
        clk       : in std_logic;

        -- Multiplexed signals
        m_awid    : in std_logic_vector(1 downto 0);
        m_awaddr  : in std_logic_vector(15 downto 0);
        m_awlen   : in std_logic_vector(1 downto 0);
        m_awsize  : in std_logic;
        m_awburst : in std_logic;
        m_awlock  : in std_logic;
        m_awcache : in std_logic_vector(1 downto 0);
        m_awprot  : in std_logic_vector(1 downto 0);
        m_awvalid : in std_logic;

        m_wid     : in std_logic_vector(1 downto 0);
        m_wdata   : in std_logic_vector(31 downto 0);
        m_wstrb   : in std_logic_vector(3 downto 0);
        m_wlast   : in std_logic;
        m_wvalid  : in std_logic;

        m_bid     : out std_logic_vector(2 downto 0);
        m_bresp   : out std_logic;
        m_bvalid  : out std_logic;

        m_arid    : in std_logic_vector(1 downto 0);
        m_araddr  : in std_logic_vector(15 downto 0);
        m_arlen   : in std_logic_vector(1 downto 0);
        m_arsize  : in std_logic;
        m_arburst : in std_logic;
        m_arlock  : in std_logic;
        m_arcache : in std_logic_vector(1 downto 0);
        m_arprot  : in std_logic_vector(1 downto 0);
        m_arvalid : in std_logic;

        m_rid     : out std_logic_vector(1 downto 0);
        m_rdata   : out std_logic_vector(31 downto 0);
        m_rresp   : out std_logic;
        m_rlast   : out std_logic;
        m_rvalid  : out std_logic;

        -- Demultiplexed signals
        aw : out axi_addr_channel;
        ar : out axi_addr_channel;
        w  : out axi_write_data_channel;
        r  : in  axi_read_data_channel;
        b  : in  axi_write_resp_channel
    );
end axi_demux;

architecture axi_demux_arch of axi_demux is
    signal awid    : std_logic_vector(5 downto 0);
    signal awaddr  : std_logic_vector(31 downto 0);
    signal awlen   : std_logic_vector(3 downto 0);
    signal awsize  : std_logic_vector(1 downto 0);
    signal awburst : std_logic_vector(1 downto 0);
    signal awlock  : std_logic_vector(1 downto 0);
    signal awcache : std_logic_vector(3 downto 0);
    signal awprot  : std_logic_vector(2 downto 0);
    signal awvalid : std_logic;

    signal wid     : std_logic_vector(5 downto 0);
    signal wdata   : std_logic_vector(63 downto 0);
    signal wstrb   : std_logic_vector(7 downto 0);
    signal wlast   : std_logic;
    signal wvalid  : std_logic;

    signal bid     : std_logic_vector(5 downto 0);
    signal bresp   : std_logic_vector(1 downto 0);
    signal bvalid  : std_logic;

    signal arid    : std_logic_vector(5 downto 0);
    signal araddr  : std_logic_vector(31 downto 0);
    signal arlen   : std_logic_vector(3 downto 0);
    signal arsize  : std_logic_vector(1 downto 0);
    signal arburst : std_logic_vector(1 downto 0);
    signal arlock  : std_logic_vector(1 downto 0);
    signal arcache : std_logic_vector(3 downto 0);
    signal arprot  : std_logic_vector(2 downto 0);
    signal arvalid : std_logic;

    signal rid     : std_logic_vector(5 downto 0);
    signal rdata   : std_logic_vector(63 downto 0);
    signal rresp   : std_logic_vector(1 downto 0);
    signal rlast   : std_logic;
    signal rvalid  : std_logic;

    signal zero    : std_logic;
begin
    zero <= '0';

    aw.id    <= awid;
    aw.addr  <= awaddr;
    aw.len   <= vector_to_axi_len(awlen);
    aw.size  <= awsize;
    aw.burst <= vector_to_axi_burst(awburst);
    aw.lock  <= vector_to_axi_lock(awlock);
    aw.cache <= vector_to_axi_cache(awcache);
    aw.prot  <= vector_to_axi_prot(awprot);
    aw.valid <= awvalid;

    w.id     <= wid;
    w.data   <= wdata;
    w.strb   <= wstrb;
    w.last   <= wlast;
    w.valid  <= wvalid;

    ar.id    <= arid;
    ar.addr  <= araddr;
    ar.len   <= vector_to_axi_len(arlen);
    ar.size  <= arsize;
    ar.burst <= vector_to_axi_burst(arburst);
    ar.lock  <= vector_to_axi_lock(arlock);
    ar.cache <= vector_to_axi_cache(arcache);
    ar.prot  <= vector_to_axi_prot(arprot);
    ar.valid <= arvalid;

    rid      <= r.id;
    rdata    <= r.data;
    rresp    <= axi_resp_to_vector(r.resp);
    rlast    <= r.last;
    rvalid   <= r.valid;

    bid      <= b.id;
    bresp    <= axi_resp_to_vector(b.resp);
    bvalid   <= b.valid;

    -- Write address channel
    AW1: entity work.demux_vector generic map(2)  port map(clk, m_awid,    awid(3 downto 0));
    AW2: entity work.demux_vector generic map(16) port map(clk, m_awaddr,  awaddr);
    AW3: entity work.demux_vector generic map(2)  port map(clk, m_awlen,   awlen);
    AW4: entity work.demux_vector generic map(2)  port map(clk, m_awcache, awcache);
    AW5: entity work.demux_vector generic map(2)  port map(clk, m_awprot,
                                              o(1) => awid(4),
                                              o(0) => awprot(0),
                                              o(3 downto 2) => awprot(2 downto 1)
                                          );
    AW6: entity work.demux_single port map(clk, m_awlock,  awlock);
    AW7: entity work.demux_single port map(clk, m_awsize,  awsize);
    AW8: entity work.demux_single port map(clk, m_awburst, awburst);
    AW9: entity work.demux_single port map(clk, m_awvalid, o(0) => awvalid, o(1) => awid(5));

    -- Read address channel
    AR1: entity work.demux_vector generic map(2)  port map(clk, m_arid,    arid(3 downto 0));
    AR2: entity work.demux_vector generic map(16) port map(clk, m_araddr,  araddr);
    AR3: entity work.demux_vector generic map(2)  port map(clk, m_arlen,   arlen);
    AR4: entity work.demux_vector generic map(2)  port map(clk, m_arcache, arcache);
    AR5: entity work.demux_vector generic map(2)  port map(clk, m_arprot,
                                              o(1) => arid(4),
                                              o(0) => arprot(0),
                                              o(3 downto 2) => arprot(2 downto 1)
                                          );
    AR6: entity work.demux_single port map(clk, m_arlock,  arlock);
    AR7: entity work.demux_single port map(clk, m_arsize,  arsize);
    AR8: entity work.demux_single port map(clk, m_arburst, arburst);
    AR9: entity work.demux_single port map(clk, m_arvalid, o(0) => arvalid, o(1) => arid(5));

    -- Write response channel
    B1: entity work.mux_vector generic map(2) port map(clk, i => bid(3 downto 0), o => m_bid(1 downto 0));
    B2: entity work.mux_single port map(clk, i(0) => bid(4), i(1) => bid(5), o => m_bid(2));
    B3: entity work.mux_single port map(clk, bresp, m_bresp);
    m_bvalid <= bvalid;

    -- Write data channel
    W1: entity work.demux_vector generic map(32) port map(clk, m_wdata, wdata);
    W2: entity work.demux_vector generic map(2)  port map(clk, m_wid,   wid(3 downto 0));
    W3: entity work.demux_vector generic map(4)  port map(clk, m_wstrb, wstrb);
    W4: entity work.demux_single port map(clk, m_wlast, o(0) => wlast, o(1) => wid(4));
    W5: entity work.demux_single port map(clk, m_wvalid, o(0) => wvalid, o(1) => wid(5));

    -- Read data channel
    R1: entity work.mux_vector generic map(32) port map(clk, rdata, m_rdata);
    R3: entity work.mux_vector generic map(2)  port map(clk, rid(3 downto 0), m_rid);
    R2: entity work.mux_single port map(clk, rresp, m_rresp);
    R4: entity work.mux_single port map(clk, i(0) => rlast, i(1) => rid(4), o => m_rlast);
    R5: entity work.mux_single port map(clk, i(0) => rvalid, i(1) => rid(5), o => m_rvalid);
end axi_demux_arch;
