-- Implements a simple AXI slave controller.
-- No support for exclusive access.
-- Ignores cache and protection signals.

-- TODO:
-- add support for fixed and wrapping bursts

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.matrix_pkg.all;
use work.axi.all;

entity axi_slave is
    port (
        aclk    : in  std_logic;
        aresetn : in  std_logic;

        aw      : in  axi_addr_channel;
        awready : out std_logic;

        w       : in  axi_write_data_channel;
        wready  : out std_logic;

        b       : out axi_write_resp_channel;
        bready  : in  std_logic;

        ar      : in  axi_addr_channel;
        arready : out std_logic;

        r       : out axi_read_data_channel;
        rready  : in  std_logic;

        matrix_in  : out matrix_master_in;
        matrix_out : in  matrix_out
    );

end axi_slave;

architecture axi_slave_arch of axi_slave is
    type write_state_type is ( ws_wait_addr
                             , ws_wait_data
                             , ws_write_data
                             , ws_incr_addr
                             , ws_send_resp
                             , ws_reset
                         );
    type read_state_type is ( rs_wait_addr
                            , rs_read_data
                            , rs_send_data
                            , rs_reset
                        );
    subtype increment is natural range 1 to 8;

    function size_to_increment(size : std_logic_vector(1 downto 0))
        return increment is
    begin
        case size is
            when b"00" => return 1;
            when b"01" => return 2;
            when b"10" => return 4;
            when b"11" => return 8;
            when others => return 8;
        end case;
    end size_to_increment;

    -- Registers for writing
    signal write_state  : write_state_type := ws_reset;
    signal write_strobe : axi_strobe;
    signal write_addr   : axi_addr;
    signal write_data   : axi_data;
    signal write_last   : std_logic;
    signal write_id     : axi_id;
    signal write_addr_incr : increment;

    -- Register for reading
    signal read_state : read_state_type := rs_reset;
    signal read_rem   : natural;
    signal read_addr  : axi_addr;
    signal read_data  : axi_data;
    signal read_id    : axi_id;
    signal read_addr_incr : increment;
begin
    -- Connection to the matrix
    matrix_in.raddr <= read_addr;
    matrix_in.ren   <= '1' when read_state = rs_read_data else '0';
    matrix_in.waddr <= write_addr;
    matrix_in.wdata <= write_data;
    matrix_in.wstrb <= write_strobe;
    read_data       <= matrix_out.rdata;


    -- Write process
    process (aclk, aresetn) is
    begin
        if aresetn = '0' then
            write_state <= ws_reset;
        elsif rising_edge(aclk) then
            -- Set write registers
            if write_state = ws_wait_addr
            then
                write_id        <= aw.id;
                write_addr      <= aw.addr;
                write_addr_incr <= size_to_increment(aw.size);
            else
                write_id        <= write_id;
                write_addr      <= write_addr;
                write_addr_incr <= write_addr_incr;
            end if;

            if write_state = ws_wait_data and w.valid = '1'
            then
                write_strobe <= w.strb;
                write_last   <= w.last;
                write_data   <= w.data;
            else
                write_strobe <= X"00";
                write_last   <= write_last;
                write_data   <= write_data;
            end if;

            -- Find the next state for writing
            case write_state is
                when ws_reset =>
                    write_state  <= ws_wait_addr;

                when ws_wait_addr =>
                    if aw.valid = '0'
                    then write_state <= ws_wait_addr;
                    else write_state <= ws_wait_data;
                    end if;

                when ws_wait_data =>
                    if w.valid = '0'
                    then write_state <= ws_wait_data;
                    else write_state <= ws_write_data;
                    end if;

                when ws_write_data =>
                    if write_last = '1'
                    then write_state <= ws_send_resp;
                    else write_state <= ws_incr_addr;
                    end if;

                when ws_incr_addr =>
                    write_addr <= std_logic_vector(unsigned(write_addr) + write_addr_incr);
                    write_state <= ws_wait_data;

                when ws_send_resp =>
                    if bready = '1'
                    then write_state <= ws_wait_addr;
                    else write_state <= ws_send_resp;
                    end if;
            end case;
        end if;
    end process;

    -- Write output signals
    b.id    <= write_id;
    b.resp  <= axi_resp_okay; -- always okay, no errors possible
    b.valid <= '1' when write_state = ws_send_resp else '0';
    awready <= '1' when write_state = ws_wait_addr else '0';
    wready  <= '1' when write_state = ws_wait_data else '0';


    -- Read process
    process (aclk, aresetn) is
    begin
        if aresetn = '0' then
            read_state <= rs_reset;
        elsif rising_edge(aclk) then
            -- Update read registers
            if read_state = rs_wait_addr
            then
                read_addr      <= ar.addr;
                read_id        <= ar.id;
                read_rem       <= ar.len;
                read_addr_incr <= size_to_increment(ar.size);
            else
                read_id        <= read_id;
                read_addr_incr <= read_addr_incr;

                if read_state = rs_read_data
                then
                    read_addr <= std_logic_vector(unsigned(read_addr) + read_addr_incr);
                    read_rem  <= read_rem - 1;
                else
                    read_addr <= read_addr;
                    read_rem  <= read_rem;
                end if;
            end if;


            -- Find next read state
            case read_state is
                when rs_reset =>
                    read_state <= rs_wait_addr;

                when rs_wait_addr =>
                    if ar.valid = '0'
                    then read_state <= rs_wait_addr;
                    else read_state <= rs_read_data;
                    end if;

                when rs_read_data =>
                    read_state <= rs_send_data;

                when rs_send_data =>
                    if rready = '1'
                    then
                        if read_rem = 0
                        then read_state <= rs_wait_addr;
                        else read_state <= rs_read_data;
                        end if;
                    else
                        read_state <= rs_send_data;
                    end if;
            end case;
        end if;
    end process;

    -- Set read output signals
    arready <= '1' when read_state = rs_wait_addr else '0';
    r.valid <= '1' when read_state = rs_send_data else '0';
    r.last  <= '1' when read_rem = 0 else '0';
    r.resp  <= axi_resp_okay;
    r.data  <= read_data;
    r.id    <= read_id;
end axi_slave_arch;
