library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package ca_pkg is
    subtype lut_type is std_logic_vector(31 downto 0);
    type ca_interface is
        record
            l : lut_type;
            o : std_logic;
        end record;

end ca_pkg;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.ca_pkg.all;

entity ca_cell is
    generic (
        lfsr_seed : integer := 0
    );
    port (
        clk       : in  std_logic;
        reset     : in  std_logic;

        n,s,e,w   : in  ca_interface;
        o         : out ca_interface;

        evolve    : in  std_logic;
        good, bad : in  std_logic
    );
end ca_cell;

architecture ca_cell_arch of ca_cell is
    signal lut_best : lut_type;
    signal lut_cur  : lut_type;

    signal idx      : unsigned(4 downto 0);
    signal c        : std_logic;
    signal lfsr_o   : std_logic_vector(31 downto 0);

    subtype lut_array_index is natural range 0 to 4;
    type lut_array_type is array (lut_array_index) of lut_type;

    function vec_to_index( v : std_logic_vector(3 downto 0)) return lut_array_index is
    begin
        case v is
            when x"0"   => return 0;
            when x"1"   => return 1;
            when x"2"   => return 2;
            when x"3"   => return 3;
            when x"4"   => return 4;
            when x"5"   => return 0;
            when x"6"   => return 1;
            when x"7"   => return 2;
            when x"8"   => return 3;
            when x"9"   => return 4;
            when x"a"   => return 0;
            when x"b"   => return 1;
            when x"c"   => return 2;
            when x"d"   => return 3;
            when x"e"   => return 4;
            when others => return 0;
        end case;
    end function;

    function crossover( lut_array : lut_array_type
                      ; rnd       : std_logic_vector(31 downto 0)
                      )
    return lut_type is
        variable a_idx : lut_array_index := vec_to_index(rnd(31 downto 28));
        variable b_idx : lut_array_index := vec_to_index(rnd(27 downto 24));
        variable point : integer range 0 to 31 := to_integer(unsigned(rnd(23 downto 19)));
        variable do_it : boolean := rnd(18 downto 13) = b"000000";
        variable ret   : lut_type;
    begin
        if do_it then
            if point = 31 then
                ret := lut_array(b_idx);
            else
                ret(31 downto point + 1) := lut_array(a_idx)(31 downto point + 1);
                ret(point downto 0)      := lut_array(b_idx)(point downto 0);
            end if;
        else
            ret := lut_array(4);
        end if;

        return ret;
    end function;

    function mutate( lut : lut_type
                   ; rnd : std_logic_vector(31 downto 0)
                   )
    return lut_type is
        variable bit_idx : integer  := to_integer(unsigned(rnd(10 downto 6)));
        variable ret     : lut_type := lut;
        variable do_it   : boolean  := (rnd(5 downto 0) = b"000000");
    begin
        if do_it then
            ret(bit_idx) := not ret(bit_idx);
        end if;

        return ret;
    end function;

begin
    idx <= ( 0 => n.o
           , 1 => s.o
           , 2 => e.o
           , 3 => w.o
           , 4 => c
           );

    o <= ( o => c
         , l => lut_best
         );

    LFSR : entity work.lfsr
    generic map (
        reset_value => lfsr_seed
    )
    port map (
        clk   => clk,
        o     => lfsr_o
    );

    process (clk, reset, lfsr_o)
        variable lut_array : lut_array_type;
        variable lut : lut_type;
    begin
        if reset = '1' then
            c <= '0';
            lut_best <= lfsr_o;
            lut_cur  <= lfsr_o;
        elsif rising_edge(clk) then
            lut_array := ( 0 => n.l
                         , 1 => s.l
                         , 2 => e.l
                         , 3 => w.l
                         , 4 => lut_best
                         );

            c  <= lut_cur(to_integer(idx));

            if evolve = '0' then
                lut_cur <= lut_best;
            elsif good = '1' then
                case lfsr_o(31 downto 28) is
                    when x"0"   => lut := n.l;
                    when x"1"   => lut := s.l;
                    when x"2"   => lut := e.l;
                    when x"3"   => lut := w.l;
                    when others => lut := lut_cur;
                end case;
                lut_best <= lut_cur;
                lut_cur  <= mutate(lut, lfsr_o);
            elsif bad = '1' then
                case lfsr_o(31 downto 28) is
                    when x"0"   => lut := n.l;
                    when x"1"   => lut := s.l;
                    when x"2"   => lut := e.l;
                    when x"3"   => lut := w.l;
                    when others => lut := lut_best;
                end case;
                lut_cur  <= mutate(lut, lfsr_o);
            end if;
        end if;
    end process;
end ca_cell_arch;
