library ieee;
use ieee.std_logic_1164.all;
use work.axi.all;

package matrix_pkg is
    constant slave_size_log : integer := 12;
    constant slave_size     : integer := 2**slave_size_log;
    constant data_width_log : integer := 3;
    constant data_width     : integer := 2**data_width_log;

    subtype matrix_data is std_logic_vector(data_width * 8 - 1 downto 0);
    subtype matrix_addr is natural range 0 to (slave_size / data_width - 1);

    type matrix_master_in is
        record
            raddr : axi_addr;
            ren   : std_logic;
            waddr : axi_addr;
            wdata : matrix_data;
            wstrb : axi_strobe;
        end record;

    type matrix_slave_in is
        record
            raddr : matrix_addr;
            ren   : std_logic;
            waddr : matrix_addr;
            wdata : matrix_data;
            wstrb : axi_strobe;
        end record;

    type matrix_out is
        record
            rdata : matrix_data;
        end record;

    type matrix_out_array is array (integer range <>) of matrix_out;
    type matrix_slave_in_array  is array (integer range <>) of matrix_slave_in;
end matrix_pkg;


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.axi.all;
use work.matrix_pkg.all;

entity matrix is
    generic (
        slave_count : integer := 1
    );
    port (
        clk   : in  std_logic;
        reset : in  std_logic;

        -- Interface to AXI
        master_in  : in  matrix_master_in;
        master_out : out matrix_out;

        -- Slave interfaces
        slaves_out : out matrix_slave_in_array(slave_count - 1 downto 0);
        slaves_in  : in  matrix_out_array(slave_count - 1 downto 0)
    );
end matrix;

architecture matrix_arch of matrix is
    subtype slave_index is natural range 0 to slave_count - 1;

    signal read_idx  : slave_index;
    signal write_idx : slave_index;
    signal read_reg  : matrix_out;
begin
    read_idx  <= to_integer(unsigned(master_in.raddr(29 downto slave_size_log))) mod slave_count;
    write_idx <= to_integer(unsigned(master_in.waddr(29 downto slave_size_log))) mod slave_count;

    GEN :
    for i in 0 to slave_count - 1 generate
        slaves_out(i).raddr <= to_integer(unsigned(master_in.raddr(slave_size_log - 1 downto data_width_log)));
        slaves_out(i).waddr <= to_integer(unsigned(master_in.waddr(slave_size_log - 1 downto data_width_log)));
        slaves_out(i).wdata <= master_in.wdata;
        slaves_out(i).ren   <= master_in.ren   when i = read_idx  else '0';
        slaves_out(i).wstrb <= master_in.wstrb when i = write_idx else x"00";
    end generate;

    process (clk)
    begin
        if rising_edge(clk) then
            if master_in.ren = '1' then
                read_reg <= slaves_in(read_idx);
            else
                read_reg <= read_reg;
            end if;
        end if;
    end process;

    master_out <= read_reg;
end architecture;
