library ieee;
use ieee.std_logic_1164.all;

package sys_regs_pkg is
    subtype sys_reg is std_logic_vector(31 downto 0);
    type sys_reg_desc is
        record
            read_only   : boolean;
            reset_value : sys_reg;
        end record;

    type sys_reg_index is
        ( SR_MAGIC
        , SR_CYCLE_COUNT
        , SR_CA_CTRL
        , SR_CA_INPUT
        , SR_CA_OUTPUT
        , SR_FITNESS_FUNC
        , SR_FITNESS_BEST
        , SR_PAD1
        );

    type desc_array is array (sys_reg_index) of sys_reg_desc;
    type sys_reg_array is array (sys_reg_index) of sys_reg;

    constant sys_reg_descs : desc_array
            := ( SR_MAGIC =>
                ( read_only   => true
                , reset_value => x"ca15c001"
                )
               , SR_CYCLE_COUNT =>
                ( read_only   => true
                , reset_value => x"00000000"
                )
               , SR_CA_CTRL =>
                ( read_only   => false
                , reset_value => x"00000000"
                )
               , SR_CA_INPUT =>
                ( read_only   => false
                , reset_value => x"00000000"
                )
               , SR_CA_OUTPUT =>
                ( read_only   => false
                , reset_value => x"00000000"
                )
               , SR_FITNESS_FUNC =>
                ( read_only   => false
                , reset_value => x"00000000"
                )
               , SR_FITNESS_BEST =>
                ( read_only   => true
                , reset_value => x"00000000"
                )
               , SR_PAD1 =>
                ( read_only   => false
                , reset_value => x"00000000"
                )
               );
    constant sys_reg_count : integer := sys_reg_descs'length;

    function int_to_index ( i : natural ) return sys_reg_index;
end sys_regs_pkg;

package body sys_regs_pkg is
    function int_to_index ( i : natural ) return sys_reg_index is
    begin
        case i is
            when 0      => return SR_MAGIC;
            when 1      => return SR_CYCLE_COUNT;
            when 2      => return SR_CA_CTRL;
            when 3      => return SR_CA_INPUT;
            when 4      => return SR_CA_OUTPUT;
            when 5      => return SR_FITNESS_FUNC;
            when 6      => return SR_FITNESS_BEST;
            when others => return SR_PAD1;
        end case;
    end function;
end package body;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.matrix_pkg.all;
use work.sys_regs_pkg.all;
use work.fitness_pkg.all;

entity sys_regs is
    port (
        clk        : in  std_logic;
        reset      : in  std_logic;

        matrix_in  : in  matrix_slave_in;
        matrix_out : out matrix_out;

        ca_best    : in  natural;
        ca_func    : out fitness_function;
        ca_input   : out std_logic_vector(31 downto 0);
        ca_output  : in  std_logic_vector(7 downto 0);
        ca_use_sw  : out boolean;
        ca_evolve  : out std_logic;
        ca_reset   : out std_logic
    );
end sys_regs;

architecture sys_reg_arch of sys_regs is
    signal regs       : sys_reg_array;
    signal wen0, wen1 : std_logic;
    signal waddr0, waddr1, raddr0, raddr1 : sys_reg_index;
begin
    waddr0 <= int_to_index((matrix_in.waddr * 2)     mod sys_reg_count);
    waddr1 <= int_to_index((matrix_in.waddr * 2 + 1) mod sys_reg_count);
    raddr0 <= int_to_index((matrix_in.raddr * 2)     mod sys_reg_count);
    raddr1 <= int_to_index((matrix_in.raddr * 2 + 1) mod sys_reg_count);

    wen0 <= matrix_in.wstrb(0) or
            matrix_in.wstrb(1) or
            matrix_in.wstrb(2) or
            matrix_in.wstrb(3);
    wen1 <= matrix_in.wstrb(4) or
            matrix_in.wstrb(5) or
            matrix_in.wstrb(6) or
            matrix_in.wstrb(7);

    process (clk, reset)
    begin
        if reset = '1' then
            for i in sys_reg_descs'range loop
                regs(i) <= sys_reg_descs(i).reset_value;
            end loop;
        elsif rising_edge(clk) then
            if wen0 = '1' and (not sys_reg_descs(waddr0).read_only) then
                regs(waddr0) <= matrix_in.wdata(31 downto 0);
            end if;
            if wen1 = '1' and (not sys_reg_descs(waddr1).read_only) then
                regs(waddr1) <= matrix_in.wdata(63 downto 32);
            end if;

            regs(SR_CYCLE_COUNT)  <= sys_reg(unsigned(regs(SR_CYCLE_COUNT)) + 1);
            regs(SR_FITNESS_BEST) <= sys_reg(to_unsigned(ca_best, 32));
            regs(SR_CA_OUTPUT)(7 downto 0) <= ca_output;
        end if;
    end process;

    ca_func   <= decode_fitness_func(regs(SR_FITNESS_FUNC));
    ca_input  <= regs(SR_CA_INPUT);
    ca_evolve <= regs(SR_CA_CTRL)(0);
    ca_use_sw <= regs(SR_CA_CTRL)(1) = '1';
    ca_reset  <= regs(SR_CA_CTRL)(2) or reset;

    matrix_out.rdata <= regs(raddr1) & regs(raddr0);
end architecture;
