#include "ca.h"
#include <stdlib.h>
#include <string.h>

struct ca_St {
	unsigned int w, h;
	lut_t *luts;
	uint8_t *prev_state;
	uint8_t *cur_state;
};

#define LUT(ca, x, y) ((ca)->luts[(x) + (y) * (ca)->w])
#define PREV_STATE(ca, x, y) ((ca)->prev_state[(x) + (y) * ((ca)->w + 2)])
#define CUR_STATE(ca, x, y) ((ca)->cur_state[(x) + (y) * ((ca)->w + 2)])

ca_t *
ca_new(int w, int h)
{
	ca_t *ca = malloc(sizeof(ca_t));

	ca->w = w;
	ca->h = h;
	ca->luts = malloc(sizeof(lut_t) * w * h);
	ca->prev_state = malloc(sizeof(uint8_t) * (w + 2) * h * 2);
	ca->cur_state = malloc(sizeof(uint8_t) * (w + 2) * h * 2);

	return ca;
}

void
ca_set_luts(ca_t *ca, const lut_t *luts)
{
	memcpy(ca->luts, luts, ca->w * ca->h * sizeof(lut_t));
}

uint32_t *
ca_run(ca_t *ca, uint32_t input, int iterations, uint8_t **states)
{
	return ca_run_initial(ca, input, iterations, states, 0);
}

uint32_t *
ca_run_initial(ca_t *ca, uint32_t input, int iterations, uint8_t **states, uint8_t initial)
{
	uint32_t array[iterations];

	for (int i = 0; i < iterations; i++) {
		array[i] = input;
	}

	return ca_run_array(ca, array, iterations, states, initial);
}

uint32_t *
ca_run_array(ca_t *ca, uint32_t *input, int iterations, uint8_t **states, uint8_t initial)
{
	uint32_t *outs = malloc(sizeof(uint32_t) * iterations);
	uint8_t *brds = NULL;

	if (states != NULL) {
		*states = brds = malloc(sizeof(uint8_t) * (ca->w + 2) * ca->h * iterations);
	}

	for (int y = 0; y < ca->h; y++) {
		for (int x = 1; x <= ca->w + 1; x++) {
			CUR_STATE (ca, x, y) = 0;
			PREV_STATE(ca, x, y) = 0;
		}

		/*
		CUR_STATE (ca, 0, y) = (input >> y) & 1;
		PREV_STATE(ca, 0, y) = (input >> y) & 1;
		*/
		PREV_STATE(ca, ca->w + 1, y) = (initial >> y) & 1;
		CUR_STATE(ca, ca->w + 1, y) = (initial >> y) & 1;
	}

	for (int it = 0; it < iterations; it++) {
		uint8_t *tmp = ca->cur_state;

		for (int y = 0; y < ca->h; y++) {
			PREV_STATE (ca, 0, y) = (input[it] >> y) & 1;
		}

		for (unsigned int y = 0; y < ca->h; y++) {
			for (unsigned int x = 1; x <= ca->w; x++) {
				uint8_t n, s, w, e, c;
				int idx;

				n = PREV_STATE(ca, x, (y - 1) % ca->h);
				s = PREV_STATE(ca, x, (y + 1) % ca->h);
				e = PREV_STATE(ca, x - 1, y);
				w = PREV_STATE(ca, x + 1, y);
				c = PREV_STATE(ca, x, y);
				idx = (n << 4) | (s << 3) | (e << 2) | (w << 1) | c;

				CUR_STATE(ca, x, y) = (LUT(ca, x - 1, y) >> idx) & 1;
			}
		}

		if (brds != NULL) {
			memcpy(brds + (ca->w + 2) * ca->h * it, ca->cur_state, (ca->w + 2) * ca->h);
		}

		outs[it] = 0;
		for (int y = ca->h - 1; y >= 0; y--) {
			outs[it] <<= 1;
			outs[it] |= CUR_STATE(ca, ca->w, y);
		}
		ca->cur_state = ca->prev_state;
		ca->prev_state = tmp;
	}

	return outs;
}

#include <curses.h>
#include "fft.h"
#define STATE(x, y, it, w, h) (states[(x) + (y) * (w + 2) + (it) * (h) * (w + 2)])

#define WIN_SIZE 16


static void
print_spectrum(double *spec, int x, int y)
{
	double max = 0;
	double limit = 1;

	for (int i = 1; i < WIN_SIZE / 2 + 1; i++) {
		if (spec[i] > max) {
			max = spec[i];
		}
	}

	while (limit < max) {
		limit *= 2;
	}

	for (int i = 1; i < WIN_SIZE / 2 + 1; i++) {
		for (int j = 0; j < 10; j++) {
			if ((limit / 10 * (j + 1)) <= spec[i]) {
				mvaddch(y + 10 - j, x + i, '#');
			} else {
				mvaddch(y + 10 - j, x + i, ' ');
			}
		}
	}
}

static void
print_out_values(int x, int y, int w, int h, int it, uint8_t *states)
{
	char buf[128];

	for (int i = 0; i < 64; i++) {
		int bits = 0;
		for (int j = 0; j < h; j++) {
			char c;
			if (it - i < 0) {
				c = ' ';
			} else {
				c = STATE(w, j, it - i, w, h) ? '#' : ' ';
				bits += STATE(w, j, it - i, w, h) ? 1 : 0;
			}
			mvaddch(y + i, x + j, c);
		}

		sprintf(buf, "%1i", bits);
		mvaddstr(y + i, x + h + 2, buf);
	}
}

void
ca_fft_init()
{
	fft_init(WIN_SIZE);
}

void
ca_anim(uint8_t *states, int w, int h, int iterations)
{
	char buf[128];
	int state_size = (w + 2) * h;
	double in_win[WIN_SIZE] = {0};
	double out_win[WIN_SIZE] = {0};
	bool run = true;

	initscr();
	cbreak();
	keypad(stdscr, TRUE);
	nodelay(stdscr, TRUE);
	noecho();

	for (int i = 0; i < iterations; i++) {
		sprintf(buf, "%3i", i);
		mvaddstr(0, 0, buf);

		for (int x = 0; x < w + 3; x++) {
			mvaddch(1, x, '-');
			mvaddch(h + 2, x, '-');
		}

		for (int y = 0; y < h; y++) {
			mvaddch(y + 2, 0, '|');
			mvaddch(y + 2, w + 2, '|');

			for (int x = 0; x < w + 1; x++) {
				mvaddch(y + 2, x + 1, STATE(x, y, i, w, h) ? '#' : ' ');
			}
		}

		double out_bits = 0;
		double in_bits = 0;

		for (int y = 0; y < h; y++) {
			out_bits += STATE(w, y, i, w, h) ? 1 : 0;
			in_bits  += STATE(0, y, i, w, h) ? 1 : 0;
		}

		memmove(out_win, out_win + 1, sizeof(double) * (WIN_SIZE - 1));
		out_win[WIN_SIZE - 1] = out_bits;

		memmove(in_win, in_win + 1, sizeof(double) * (WIN_SIZE - 1));
		in_win[WIN_SIZE - 1] = in_bits;

		double *out_spec = fft_run(out_win);
		for (int j = 0; j < WIN_SIZE / 2 + 1; j++) {
			sprintf(buf, "%2i - %3.2f", j, out_spec[j]);
			mvaddstr(h + 5 + j, 0, buf);
		}
		print_spectrum(out_spec, 0, h + 20);


		double *in_spec = fft_run(in_win);
		for (int j = 0; j < WIN_SIZE / 2 + 1; j++) {
			sprintf(buf, "| %2i - %3.2f", j, in_spec[j]);
			mvaddstr(h + 5 + j, 12, buf);
		}
		print_spectrum(out_spec, 0, h + 32);

		print_out_values(40, 2, w,  h, i, states);
		refresh();

		do {
			int ch = getch();

			if (ch == ' ') {
				run = !run;
			} else if (ch != ERR) {
				break;
			}
			usleep(100 * 1000);
		} while (!run);
	}
	endwin();
}

void
ca_dump(FILE *f, int prefix, uint8_t *states, int w, int h, int iterations)
{
	int outs[iterations];
	double fft_in[WIN_SIZE];
	double *spec;

	fprintf(f, "\n\n");
	fprintf(f, "# %i.out-img\n", prefix);
	for (int i = 0; i < iterations; i++) {
		for (int y = 0; y < h; y++) {
			fprintf(f, "%i %i %i\n", i, y, STATE(w, y, i, w, h) ? 1 : 0);
		}
	}

	fprintf(f, "\n\n");
	fprintf(f, "# %i.in-img\n", prefix);
	for (int i = 0; i < iterations; i++) {
		for (int y = 0; y < h; y++) {
			fprintf(f, "%i %i %i\n", i, y + 20, STATE(0, y, i, w, h) ? 1 : 0);
		}
	}

	fprintf(f, "\n\n");
	fprintf(f, "# %i.out-val\n", prefix);
	for (int i = 0; i < iterations; i++) {
		int value = 0;
		for (int y = 0; y < h; y++) {
			value += STATE(w, y, i, w, h) ? 1 : 0;
		}
		fprintf(f, "%i %i\n", i, value + 10);
		outs[i] = value;
	}

	fprintf(f, "\n\n");
	fprintf(f, "# %i.in-val\n", prefix);
	for (int i = 0; i < iterations; i++) {
		int value = 0;
		for (int y = 0; y < h; y++) {
			value += STATE(0, y, i, w, h) ? 1 : 0;
		}
		fprintf(f, "%i %i\n", i, value + 30);
	}

	for (int i = 0; i < WIN_SIZE; i++) {
		fft_in[i] = outs[i + WIN_SIZE - 1];
	}

	fprintf(f, "\n\n# %i.avgs\n", prefix);
	for (int j = WIN_SIZE; j < iterations - WIN_SIZE; j++) {
		memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
		fft_in[WIN_SIZE - 1] = outs[j + WIN_SIZE - 1];

		spec = fft_run(fft_in);
		fprintf(f, "%i %f\n", prefix, spec_avg(spec, WIN_SIZE / 2 + 1));
	}
}
