#include <complex.h>
#include <fftw3.h>
#include <math.h>

static fftw_plan fft_plan;
static complex *fft_out;
static double *fft_ret;
static int fft_width;

static void
fft_init(int w)
{
	double *in = malloc(sizeof(double) * w);
	fft_out = malloc(sizeof(complex) * (w / 2 + 1));
	fft_ret = malloc(sizeof(double) * (w / 2 + 1));

	fft_width = w / 2 + 1;

	fft_plan = fftw_plan_dft_r2c_1d(w, in, fft_out, 0);
}

static double *
fft_run(double *nums)
{
	fftw_execute_dft_r2c(fft_plan, nums, fft_out);

	for (int i = 0; i < fft_width; i++) {
		fft_ret[i] = cabs(fft_out[i]);
	}

	return fft_ret;
}

static int
spec_max(const double *spec, int len)
{
	int idx = 0;
	int max = spec[0];

	for (int i = 1; i < len; i++) {
		if (max < spec[i]) {
			idx = i;
			max = spec[i];
		} else if (max == spec[i]) {
			idx = -1;
		}
	}

	return idx;
}

static double
spec_avg(const double *spec, int len)
{
	double weighted = 0;
	double tot = 0;

	for (int i = 1; i < len; i++) {
		weighted += i * spec[i];
		tot += spec[i];
	}

	return tot == 0 ? (len) / 2.0 : weighted / tot;
}

static double
spec_var(const double *spec, int len)
{
	double mean = 0;
	double var = 0;
	double tot = 0;

	for (int i = 1; i < len; i++) {
		mean += i * spec[i];
		var += i * i * spec[i];
		tot += spec[i];
	}

	if (tot == 0) {
		return 1e100;
	} else {
		mean /= tot;
		var /= tot;

		return var - mean * mean;
	}
}


static double
array_avg(const double *arr, int size)
{
	double tot = 0;

	for (int i = 0; i < size; i++) {
		tot += arr[i];
	}

	return tot / size;
}

static double
array_span(const double *arr, int size)
{
	double min = arr[0];
	double max = arr[0];

	for (int i = 0; i < size; i++) {
		double x = arr[i];

		if (x < min) {
			min = x;
		}
		if (x > max) {
			max = x;
		}
	}

	return max - min;
}

static double
max(double a, double b)
{
	return a > b ? a : b;
}
static double
min(double a, double b)
{
	return a < b ? a : b;
}

static double
intra_cluster_distance(const double *cluster, int size)
{
	double lo = cluster[0];
	double hi = cluster[0];

	for (int i = 0; i < size; i++) {
		lo = min(lo, cluster[i]);
		hi = max(hi, cluster[i]);
	}

	return hi - lo;
}

static double
inter_cluster_distance(const double *a, const double *b, int size_a, int size_b)
{
	return fabs(array_avg(a, size_a) - array_avg(b, size_b));
}

static double
dunn_index(double **clusters, int cluster_count, int *cluster_sizes)
{
	double max_intra_cluster = 0;
	double min_inter_cluster = 1e100;

	for (int k = 0; k < cluster_count; k++) {
		max_intra_cluster = max(max_intra_cluster, intra_cluster_distance(clusters[k], cluster_sizes[k]));
	}

	for (int i = 0; i < cluster_count; i++) {
		for (int j = i + 1; j < cluster_count; j++) {
			min_inter_cluster = min(min_inter_cluster, inter_cluster_distance(clusters[i], clusters[j],
																			  cluster_sizes[i], cluster_sizes[j]));
		}
	}
	max_intra_cluster += 0.001;

	/*
	printf("max_intra: %.100f\n", max_intra_cluster);
	printf("min_inter: %.100f\n", min_inter_cluster);
	printf("ret:       %.100f\n", min_inter_cluster / max_intra_cluster);
	*/

	return max_intra_cluster == 0 ? 0 : min_inter_cluster / max_intra_cluster;
}

static double
linear_regression(double **clusters, int cluster_count, int *cluster_sizes)
{
	double beta = 0;

	/*
	double mean_x = cluster_count * (cluster_count + 1) / 2;
	double mean_x_sq = mean_x * (2 * cluster_count + 1) / 3;
	double mean_y = 0;
	double mean_xy;

	for (int i = 0; i < cluster_count; i++) {
		double avg = array_avg(clusters[i], cluster_size);
		mean_y += avg;
		mean_xy += (i + 1) * avg;
	}
	mean_y /= cluster_count;
	mean_xy /= cluster_count;

	beta = -(mean_xy - mean_x * mean_y) / (mean_x_sq - mean_x * mean_x);
	*/
	double mean_x = cluster_count * (cluster_count + 1) / 2;
	double mean_y = 0;
	double a = 0;
	double b = 0;

	for (int i = 0; i < cluster_count; i++) {
		mean_y += array_avg(clusters[i], cluster_sizes[i]);
	}

	mean_y /= cluster_count;

	for (int i = 0; i < cluster_count; i++) {
		a += (i + 1 - mean_x) * (array_avg(clusters[i], cluster_sizes[i]) - mean_y);
		b += (i + 1 - mean_x) * (i + 1 - mean_x);
	}

	beta = a / b;

	return dunn_index(clusters, cluster_count, cluster_sizes) * beta;
}

static double
berg_index(double **clusters, int cluster_count, int *cluster_sizes)
{
	double outside = 0;
	double c_dist, c_span;

	for (int a = 0; a < cluster_count; a++) {
		double max_a, min_a;

		max_a = clusters[a][0];
		min_a = clusters[a][0];

		for (int i = 0; i < cluster_sizes[a]; i++) {
			double x = clusters[a][i];
			if (x > max_a)
				max_a = x;
			if (x < min_a)
				min_a = x;
		}

		for (int b = 0; b < cluster_count; b++) {
			if (b == a)
				continue;

			for (int i = 0; i < cluster_sizes[b]; i++) {
				double x = clusters[b][i];

				if (x < min_a || x > max_a)
					outside++;
			}
		}
	}

//	c_dist = fabs(array_avg(clusters[0], cluster_size) - array_avg(clusters[1], cluster_size));
//	c_span = fmax(array_span(clusters[0], cluster_size), array_span(clusters[1], cluster_size));

	return outside; // c_dist / fmax(0.5, c_span); //outside * 10 + c_dist - c_span;
}
