#ifndef WIN_SIZE
#define WIN_SIZE 16
#endif

typedef struct {
	lut_t lut;
	uint8_t initial;
} uni_genome_t;

static int
count_bits(uint32_t b)
{
	int bits = 0;
	for (; b; b >>= 1) {
		bits += b & 1;
	}
	return bits;
}

static void
_srand(void)
{
	srand48(time(NULL));
	srand(time(NULL));
	/*
	srand48(0);
	srand(0);
	*/
}

static uint32_t
_rand(void)
{
	return (uint32_t)mrand48();
}

static void *
uni_new(void)
{
	uni_genome_t *ret = malloc(sizeof(uni_genome_t));

	ret->lut = _rand();
	ret->initial = _rand();

	return ret;
}

static void *
uni_mutate(void *g)
{
	uni_genome_t *gen = malloc(sizeof(uni_genome_t));
	memcpy(gen, g, sizeof(uni_genome_t));

	for (int i = 0; i < 5; i++) {
		int b = _rand() % 32;

		gen->lut ^= 1 << b;

		if (rand() % 2 != 0) {
			break;
		}
	}

	for (int i = 0; i < 5; i++) {
		if (rand() % 5 != 0) {
			break;
		}

		int b = rand() % 8;

		gen->initial ^= 1 << b;
	}

	return gen;
}

static double
uni_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	uni_genome_t *gen = g;
	int tot = 0;
	lut_t luts[CA_W * CA_H];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->lut;
	}

	ca_set_luts(ca, luts);

	for (int i = 0; i < 16; i++) {
		uint32_t in = i * 15;// + _rand() % 16;
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			tot += FIT_FUNC(outs[j], in);
		}

		free(outs);
	}

	return tot;
}

static void
uni_crossover(void *a, void *b, void **c, void **d)
{
	uni_genome_t *ga = a;
	uni_genome_t *gb = b;
	uni_genome_t *gc = malloc(sizeof(uni_genome_t));
	uni_genome_t *gd = malloc(sizeof(uni_genome_t));

	gc->lut = 0;
	gd->lut = 0;

	for (int i = 31; i >= 0; i--) {
		gc->lut <<= 1;
		gd->lut <<= 1;

		if (rand() % 2) {
			gc->lut |= (ga->lut >> i) & 1;
			gd->lut |= (gb->lut >> i) & 1;
		} else {
			gd->lut |= (ga->lut >> i) & 1;
			gc->lut |= (gb->lut >> i) & 1;
		}
	}

	gc->initial = 0;
	gd->initial = 0;
	for (int i = 7; i >= 0; i--) {
		gc->initial <<= 1;
		gd->initial <<= 1;

		if (rand() % 2) {
			gc->initial |= (ga->initial >> i) & 1;
			gd->initial |= (gb->initial >> i) & 1;
		} else {
			gd->initial |= (ga->initial >> i) & 1;
			gc->initial |= (gb->initial >> i) & 1;
		}
	}

	*c = gc;
	*d = gd;
}

static void*
uni_copy(void *g)
{
	uni_genome_t *c = malloc(sizeof(uni_genome_t));
	memcpy(c, g, sizeof(uni_genome_t));
	return c;
}

ea_cfg_t uni_cfg =
	{ .new = uni_new
	, .copy = uni_copy
	, .mutate = uni_mutate
	, .crossover = uni_crossover
	, .calc_fitness = uni_fitness
	, .user_data = NULL
	, .mut_rate = 0.7
	, .cross_rate = 0.0
	};



typedef struct {
	lut_t luts[CA_W * CA_H];
	uint8_t initial;
} nu_genome_t;

static void *
nu_new(void)
{
	nu_genome_t *ret = malloc(sizeof(nu_genome_t));

	for (int y = 0; y < CA_H; y++) {
		for (int x = 0; x < CA_W; x++) {
			ret->luts[x + y * CA_W] = _rand();
			//ret->luts[x + y * CA_W] = 0xf0f0f0f0;
		}
	}
	ret->initial = _rand();

	return ret;
}

static void *
nu_mutate(void *g)
{
	nu_genome_t *gen = malloc(sizeof(nu_genome_t));
	memcpy(gen, g, sizeof(nu_genome_t));

	for (int i = 0; i < 5; i++) {
		unsigned int x = _rand() % CA_W;
		unsigned int y = _rand() % CA_H;

		switch (rand() % 8) {
		case 0:
			gen->luts[x + y * CA_W] = gen->luts[((x + 1) % CA_W) + ((y + 1) % CA_H) * CA_W];
			break;
		case 1:
			gen->luts[x + y * CA_W] = gen->luts[((x + 1) % CA_W) + ((y - 1) % CA_H) * CA_W];
			break;
		case 2:
			gen->luts[x + y * CA_W] = gen->luts[((x - 1) % CA_W) + ((y + 1) % CA_H) * CA_W];
			break;
		case 3:
			gen->luts[x + y * CA_W] = gen->luts[((x - 1) % CA_W) + ((y - 1) % CA_H) * CA_W];
			break;
		default:
			gen->luts[x + y * CA_W] = _rand();
			break;
		}

		if (rand() % 2 != 0) {
			break;
		}
	}

	for (int i = 0; i < 5; i++) {
		if (rand() % 5 != 0) {
			break;
		}

		int b = rand() % 8;

		gen->initial ^= 1 << b;
	}

	return gen;
}

static void
nu_crossover(void *a, void *b, void **c, void **d)
{
	nu_genome_t *ga = a;
	nu_genome_t *gb = b;
	nu_genome_t *gc = malloc(sizeof(nu_genome_t));
	nu_genome_t *gd = malloc(sizeof(nu_genome_t));

	for (int i = 0; i < CA_W * CA_H; i++) {
		if (rand() % 2) {
			gc->luts[i] = ga->luts[i];
			gd->luts[i] = gb->luts[i];
		} else {
			gd->luts[i] = ga->luts[i];
			gc->luts[i] = gb->luts[i];
		}
	}
	gc->initial = 0;
	gd->initial = 0;
	for (int i = 7; i >= 0; i--) {
		gc->initial <<= 1;
		gd->initial <<= 1;

		if (rand() % 2) {
			gc->initial |= (ga->initial >> i) & 1;
			gd->initial |= (gb->initial >> i) & 1;
		} else {
			gd->initial |= (ga->initial >> i) & 1;
			gc->initial |= (gb->initial >> i) & 1;
		}
	}


	*c = gc;
	*d = gd;
}

static double
nu_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	nu_genome_t *gen = g;
	double tot = 0;

	ca_set_luts(ca, gen->luts);


	for (int i = 0; i < 16; i++) {
		uint32_t in = i * 15;// + _rand() % 16;
		//uint32_t in = _rand() & 0xff;
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			tot += FIT_FUNC(outs[j], in);
			/*
			if (FIT_FUNC(outs[j], in)) {
				tot++;
			}
			*/
		}

		free(outs);
	}

	return tot;
}

static void*
nu_copy(void *g)
{
	nu_genome_t *c = malloc(sizeof(nu_genome_t));
	memcpy(c, g, sizeof(nu_genome_t));
	return c;
}

static void
nu_dump(const char *filename, nu_genome_t *gen, int w)
{
	FILE *f;

	f = fopen(filename, "w");

	fprintf(f, "%i %02x\n", w, gen->initial);
	for (int y = 0; y < CA_H; y++) {
		for (int x = 0; x < w; x++) {
			fprintf(f, "%08x ", gen->luts[x + y * CA_W]);
		}
		fprintf(f, "\n");
	}

	fclose(f);
}

ea_cfg_t nu_cfg =
	{ .new = nu_new
	, .copy = nu_copy
	, .mutate = nu_mutate
	, .crossover = nu_crossover
	, .calc_fitness = nu_fitness
	, .user_data = NULL
	, .mut_rate = 0.2
	, .cross_rate = 0.5
	};


static int cell_count = 1;

typedef struct {
	lut_t luts[10];
	uint8_t initial;
} su_genome_t;

static void *
su_new(void)
{
	su_genome_t *ret = malloc(sizeof(su_genome_t));

	for (int i = 0; i < cell_count; i++) {
		ret->luts[i] = _rand();
	}
	ret->initial = _rand();

	return ret;
}

static void *
su_mutate(void *g)
{
	su_genome_t *gen = malloc(sizeof(su_genome_t));
	memcpy(gen, g, sizeof(su_genome_t));

	switch (rand() % 4) {
	case 0:
	case 1:
	case 2:
		for (int i = 0; i < 32; i++) {
			int c = rand() % cell_count;
			int b = rand() % 32;

			gen->luts[c] ^= 1 << b;

			if (rand() % 2 != 0) {
				break;
			}
		}

	case 3:
		for (int i = 0; i < 8; i++) {
			int b = rand() % 8;

			gen->initial ^= 1 << b;

			if (rand() % 2 != 0) {
				break;
			}
		}
	}

	return gen;
}

static void
su_crossover(void *a, void *b, void **c, void **d)
{
	su_genome_t *ga = a;
	su_genome_t *gb = b;
	su_genome_t *gc = malloc(sizeof(su_genome_t));
	su_genome_t *gd = malloc(sizeof(su_genome_t));

	for (int i = 0; i < cell_count; i++) {
		if (rand() % 2) {
			gc->luts[i] = ga->luts[i];
			gd->luts[i] = gb->luts[i];
		} else {
			gd->luts[i] = ga->luts[i];
			gc->luts[i] = gb->luts[i];
		}
	}
	gc->initial = 0;
	gd->initial = 0;
	for (int i = 7; i >= 0; i--) {
		gc->initial <<= 1;
		gd->initial <<= 1;

		if (rand() % 2) {
			gc->initial |= (ga->initial >> i) & 1;
			gd->initial |= (gb->initial >> i) & 1;
		} else {
			gd->initial |= (ga->initial >> i) & 1;
			gc->initial |= (gb->initial >> i) & 1;
		}
	}


	*c = gc;
	*d = gd;
}

static void*
su_copy(void *g)
{
	su_genome_t *c = malloc(sizeof(su_genome_t));
	memcpy(c, g, sizeof(su_genome_t));
	return c;
}

static void
su_dump(FILE *f, su_genome_t *gen)
{
	fprintf(f, "%i ", cell_count);

	for (int i = 0; i < cell_count; i++) {
		fprintf(f, "%08x ", gen->luts[i]);
	}

	fprintf(f, "%02x\n", gen->initial);
}

static su_genome_t *
su_read(FILE *f)
{
	su_genome_t *gen = malloc(sizeof(su_genome_t));

	fscanf(f, "%i", &cell_count);

	for (int i = 0; i < cell_count; i++) {
		fscanf(f, "%x", gen->luts + i);
	}
	fscanf(f, "%x", &gen->initial);

	return gen;
}

ea_cfg_t su_cfg =
	{ .new = su_new
	, .copy = su_copy
	, .mutate = su_mutate
	, .crossover = su_crossover
	, .calc_fitness = NULL
	, .user_data = NULL
	, .mut_rate = 0.2
	, .cross_rate = 0.5
	};


int ls_types = 2;

#define LS_TYPES ls_types
#define LS_SUCCS 128

typedef struct {
	lut_t luts[10];
	uint8_t initial;

	struct {
		int len;
		int succ[LS_SUCCS];
	} rules[10];
} ls_genome_t;

static void
ls_replace_sym(int pos, int *str, int *len)
{
	int new_str[LS_SUCCS + 10];
	int new_len = 0;
	int new_pos;

	if (str[pos] == -4) {
		pos--;
	}

	if (pos >= LS_SUCCS) {
		abort();
	}

	do {
		int brackets = 0;
		int old_pos = pos;

		/* Copy first part */
		new_pos = pos;
		memcpy(new_str, str, old_pos * sizeof(int));

		/* Replace symbol with new string */
		int syms = 1 + (rand() % 5);

		if (rand() % 3 == 0) {
			brackets = 1;
			new_str[new_pos++] = -3;
		}

#ifdef GEN
		switch (rand() % 4) {
		case 1:
			new_str[new_pos++] = -1;
			break;
		case 2:
			new_str[new_pos++] = -2;
		}
#endif

		for (int s = 0; s < syms; s++) {
#ifdef GEN
			int c = rand() % (LS_TYPES);
#else
			int c = (rand() % (LS_TYPES + 2)) - 2;
#endif
			new_str[new_pos++] = c;
		}

		for (; brackets > 0; brackets--) {
			new_str[new_pos++] = -4;
		}

		/* Copy last part */
		if (str[old_pos] == -3) {
			int brackets = 1;
			old_pos++;

			for (; old_pos < *len && brackets > 0; old_pos++) {
				if (str[old_pos] == -3) {
					brackets++;
				} else if (str[old_pos] == -4) {
					brackets--;
				}
			}
		} else {
			old_pos++;
		}

		memcpy(new_str + new_pos, str + old_pos, (*len - old_pos) * sizeof(int));
		new_len = new_pos + *len - old_pos;
	} while (new_len > LS_SUCCS);

	*len = new_len;
	memcpy(str, new_str, new_len * sizeof(int));
}

static void *
ls_new(void)
{
	ls_genome_t *ret = malloc(sizeof(ls_genome_t));

	for (int i = 0; i < LS_TYPES; i++) {
		ret->luts[i] = _rand();

		ret->rules[i].succ[0] = i;
		ret->rules[i].len = 1;
//		ls_replace_sym(0, ret->rules[i].succ, &ret->rules[i].len);
	}
	ret->initial = _rand();

	return ret;
}

static int
ls_size(ls_genome_t *gen)
{
	int s = 0;

	for (int i = 0; i < LS_TYPES; i++) {
		s += gen->rules[i].len;
	}

	return s;
}

static void *
ls_mutate(void *g)
{
	ls_genome_t *gen = malloc(sizeof(ls_genome_t));
	memcpy(gen, g, sizeof(ls_genome_t));

	switch (rand() % 4) {
	case 0:
	case 1:
		for (int i = 0; i < 32; i++) {
			int c = rand() % LS_TYPES;
			int b = rand() % 32;

			gen->luts[c] ^= 1 << b;

			if (rand() % 2 != 0) {
				break;
			}
		}
		break;

	case 2:
		for (int i = 0; i < 8; i++) {
			int b = rand() % 8;

			gen->initial ^= 1 << b;

			if (rand() % 2 != 0) {
				break;
			}
		}
		break;

	case 3:
		for (int i = 0; i < LS_TYPES; i++) {
			int r = rand() % LS_TYPES;
			int s = rand() % gen->rules[r].len;

			ls_replace_sym(s, gen->rules[r].succ, &gen->rules[r].len);

			if (rand() % 2 != 0) {
				break;
			}
		}
		break;
	}

	return gen;
}

static void
ls_crossover(void *a, void *b, void **c, void **d)
{
	ls_genome_t *ga = a;
	ls_genome_t *gb = b;
	ls_genome_t *gc = malloc(sizeof(ls_genome_t));
	ls_genome_t *gd = malloc(sizeof(ls_genome_t));

	int cp = rand() % LS_TYPES;

	for (int i = 0; i < LS_TYPES; i++) {
		if (i < cp) {
			gc->luts[i] = ga->luts[i];
			gd->luts[i] = gb->luts[i];
		} else {
			gd->luts[i] = ga->luts[i];
			gc->luts[i] = gb->luts[i];
		}
	}

	gc->initial = ga->initial;
	gd->initial = gb->initial;

	/*
	for (int i = 7; i >= 0; i--) {
		gc->initial <<= 1;
		gd->initial <<= 1;

		if (rand() % 2) {
			gc->initial |= (ga->initial >> i) & 1;
			gd->initial |= (gb->initial >> i) & 1;
		} else {
			gd->initial |= (ga->initial >> i) & 1;
			gc->initial |= (gb->initial >> i) & 1;
		}
	}
	*/

	for (int i = 0; i < LS_TYPES; i++) {
		if (i < cp) {
			gc->rules[i] = ga->rules[i];
			gd->rules[i] = gb->rules[i];
		} else {
			gd->rules[i] = ga->rules[i];
			gc->rules[i] = gb->rules[i];
		}
	}

	*c = gc;
	*d = gd;
}

static void*
ls_copy(void *g)
{
	ls_genome_t *c = malloc(sizeof(ls_genome_t));
	memcpy(c, g, sizeof(ls_genome_t));
	return c;
}

/*
static void
ls_run(ls_genome_t *gen, int *out_array, int w, int h)
{
	int *old, *new, *tmp;
	int array[w * h * 2];
	int len = 1;
	int it;

	old = array;
	new = array + w * h;

	memset(array, 0, w * h * 2 * sizeof(int));

	for (it = 0; (len < w * h) && (it < w * h); it++) {
		int old_pos = 0;
		int new_pos = 0;

		for (; old_pos < len && new_pos < w * h; old_pos++) {
			int c = old[old_pos];

			for (int i = 0; i < gen->rules[c].len && new_pos < w * h; i++) {
				new[new_pos++] = gen->rules[c].succ[i];
			}
		}

		len = new_pos;

		tmp = old;
		old = new;
		new = tmp;
	}

	for (int i = 0; i < w * h; i++) {
		out_array[i] = old[i];
	}
}
*/

static int
ls_paint(int *prog, int pos, int plen, int *canvas, int x, int y, int dx, int dy, int w, int h)
{
	int tmp;

	for (int i = pos; i < plen; i++) {
		switch (prog[i]) {
		case -1:
			tmp = dx;
			dx = dy;
			dy = -tmp;
			break;

		case -2:
			tmp = dx;
			dx = -dy;
			dy = tmp;
			break;

		case -3:
			i = ls_paint(prog, i + 1, plen, canvas, x, y, dx, dy, w, h);
			break;

		case -4:
			return i;

		default:
			if (prog[i] < 0 || prog[i] >= LS_TYPES) {
				abort();
			}
			canvas[x + y * w] = prog[i];
			x = (x + dx + w) % w;
			y = (y + dy + h) % h;
			break;
		}
	}

	return plen;
}

#define ARR_LEN 16384

static void
ls_run(ls_genome_t *gen, int *out_array, int w, int h)
{
	int *old, *new, *tmp;
	int array[ARR_LEN * 2];
	int len = 1;
	int it;

	array[0] = LS_TYPES > 1 ? 1 : 0;

	old = array;
	new = array + ARR_LEN;

	for (it = 0; (len < ARR_LEN) && (it < 4); it++) {
		int old_pos = 0;
		int new_pos = 0;

		for (; old_pos < len && new_pos < ARR_LEN; old_pos++) {
			int c = old[old_pos];

			if (c < 0) {
				new[new_pos++] = c;
			} else {
				for (int i = 0; i < gen->rules[c].len && new_pos < ARR_LEN; i++) {
					new[new_pos++] = gen->rules[c].succ[i];
				}
			}
		}

		len = new_pos;

		tmp = old;
		old = new;
		new = tmp;
	}

	memset(out_array, 0, w * h * sizeof(int));
	ls_paint(old, 0, len, out_array, 0, 0, 1, 0, w, h);
}

static int
ls_xy_to_n(int x, int y, int w, int h)
{
	return x + y * w;
//	return y + x * h;

	/*
	int n = 0;

	x -= (w - 1) / 2;
	y -= h / 2;

	if        ( x > abs(y)) {
		n = 4 * x * x - 3 * x - y;
	} else if (-x > abs(y)) {
		n = 4 * x * x - 1 * x + y;
	} else if ( y >= abs(x)) {
		n = 4 * y * y + 3 * y + x;
	} else if (-y >= abs(x)) {
		n = 4 * y * y + 1 * y - x;
	}

	return n;
	*/
}

static char
ls_int_to_char(int c)
{
	switch (c) {
	case -1:
		return '-';
	case -2:
		return '+';
	case -3:
		return '[';
	case -4:
		return ']';
	default:
		return c + 'a';
	}
}

static int
ls_char_to_int(int c)
{
	switch (c) {
	case '-':
		return -1;
	case '+':
		return -2;
	case '[':
		return -3;
	case ']':
		return -4;
	default:
		return c - 'a';
	}
}

static void
ls_dump(FILE *f, ls_genome_t *gen, int w, int h)
{
	int array[w * h];

	fprintf(f, "%i %i %i\n", LS_TYPES, w, h);
	for (int i = 0; i < LS_TYPES; i++) {
		fprintf(f, "%08x ", gen->luts[i]);
	}

	fprintf(f, "%02x\n", gen->initial);

	for (int i = 0; i < LS_TYPES; i++) {
		fprintf(f, "%c -> ", 'a' + i);
		for (int j = 0; j < gen->rules[i].len; j++) {
			fprintf(f, "%c", ls_int_to_char(gen->rules[i].succ[j]));
		}
		fprintf(f, "\n");
	}

	ls_run(gen, array, w, h);

	for (int y = 0; y < h; y++) {
		for (int x = 0; x < w; x++) {
			int n = ls_xy_to_n(x, y, w, h);

			fprintf(f, "%c", ls_int_to_char(array[n]));
		}
		fprintf(f, "\n");
	}
}

static ls_genome_t *
ls_read(FILE *f, int *w, int *h)
{
	char buf[256];
	ls_genome_t *gen = malloc(sizeof(ls_genome_t));

	fscanf(f, "%i %i %i", &ls_types, w, h);

	for (int i = 0; i < ls_types; i++) {
		fscanf(f, "%x", &gen->luts[i]);
	}
	fscanf(f, "%x", &gen->initial);

	for (int i = 0; i < ls_types; i++) {
		fscanf(f, "%*s -> %s", buf);
		printf("buf: %s\n", buf);

		gen->rules[i].len = 0;
		for (int j = 0; buf[j] != 0; j++) {
			gen->rules[i].succ[j] = ls_char_to_int(buf[j]);
			gen->rules[i].len++;
		}
	}

	return gen;
}

static void
ls_set_luts(ca_t *ca, ls_genome_t *gen, int w, int h)
{
	lut_t luts[w * h];
	int array[w * h];

	ls_run(gen, array, w, h);

	for (int y = 0; y < h; y++) {
		for (int x = 0; x < w; x++) {
			int n = ls_xy_to_n(x, y, w, h);

			//printf("x: %i, y: %i, n: %i, w: %i, h: %i\n", x, y, n, w, h);

			luts[x + y * w] = gen->luts[array[n]];
		}
	}

	ca_set_luts(ca, luts);
}

#define MIN(a, b) ((a < b) ? a : b)

static int
levenshtein(int len_a, int *a, int len_b, int *b)
{
	int m[len_a][len_b];

	for (int i = 0; i < len_a; i++) {
		m[i][0] = i;
	}
	for (int i = 0; i < len_b; i++) {
		m[0][i] = i;
	}

	for (int i = 1; i < len_a; i++) {
		for (int j = 1; j < len_b; j++) {
			int x, y, z;

			x = m[i - 1][j];
			y = m[i][j - 1];
			z = m[i - 1][j - 1] + (a[i] == b[j] ? 1 : 0);

			m[i][j] = MIN(MIN(x, y), z);
		}
	}

	return m[len_a - 1][len_b - 1];
}

static int
simple_distance(int len_a, int *a, int len_b, int *b)
{
	int diff = abs(len_a - len_b);

	for (int i = 0; i < len_a && i < len_b; i++) {
		diff += a[i] != b[i];
	}

	return diff;
}

static double
ls_distance(void *a, void *b)
{
	ls_genome_t *ga = a;
	ls_genome_t *gb = b;
	double ret = 0;

	for (int i = 0; i < LS_TYPES; i++) {
//		ret += count_bits(ga->luts[i] ^ gb->luts[i]);
//		ret += levenshtein(ga->rules[i].len, ga->rules[i].succ, gb->rules[i].len, gb->rules[i].succ);
		ret += simple_distance(ga->rules[i].len, ga->rules[i].succ, gb->rules[i].len, gb->rules[i].succ);
	}

	return ret;
}

ea_cfg_t ls_cfg =
	{ .new = ls_new
	, .copy = ls_copy
	, .mutate = ls_mutate
	, .crossover = ls_crossover
	, .distance = ls_distance
	, .calc_fitness = NULL
	, .user_data = NULL
	, .mut_rate = 0.5
	, .cross_rate = 0.50
	};

static void *
ls2_mutate(void *g)
{
	ls_genome_t *gen = malloc(sizeof(ls_genome_t));
	memcpy(gen, g, sizeof(ls_genome_t));

	switch (rand() % 4) {
	case 0:
	case 1:
		for (int i = 0; i < 32; i++) {
			int c = rand() % LS_TYPES;
			int b = rand() % 32;

			gen->luts[c] ^= 1 << b;

			if (rand() % 2 != 0) {
				break;
			}
		}
		break;

	case 2:
		for (int i = 0; i < 8; i++) {
			int b = rand() % 8;

			gen->initial ^= 1 << b;

			if (rand() % 2 != 0) {
				break;
			}
		}
		break;

	case 3:
		for (int i = 0; i < LS_TYPES; i++) {
			int r = rand() % LS_TYPES;
//			int r = 1;
			int s = rand() % gen->rules[r].len;

			ls_replace_sym(s, gen->rules[r].succ, &gen->rules[r].len);

			if (rand() % 2 != 0) {
				break;
			}
		}
		break;
	}

	return gen;
}

static void
find_subtree(int *string, int len, int p, int *start, int *stop)
{
	int nesting = 0;
	int a, b;

	if (string[p] == -3) {
		a = p + 1;
		nesting = 1;

		for (b = p + 1; b < len; b++) {
			int c = string[b];

			if (c == -3) {
				nesting++;
			} else if (c == -4) {
				nesting--;

				if (nesting == 0) {
					break;
				}
			}
		}
	} else if (string[p] == -4) {
		b = p;
		nesting = 1;

		for (a = p -1; a > 0; a--) {
			int c = string[a];

			if (c == -3) {
				nesting--;

				if (nesting == 0) {
					a++;
					break;
				}
			} else if (c == -4) {
				nesting++;
			}
		}
	} else {
		for (a = p; a > 0; a--) {
			int c = string[a];

			if (c == -3 || c == -4) {
				a++;
				break;
			}
		}
		for (b = p + 1; b < len; b++) {
			int c = string[b];

			if (c == -3 || c == -4) {
				break;
			}
		}
	}


	*start = a;
	*stop = b;
}

static void
intersperse_subtree(int *a, int start_a, int stop_a, int len_a,
                    int *b, int start_b, int stop_b,
                    int *dest)
{
	memcpy(dest,
	       a,
	       start_a * sizeof(int));
	memcpy(dest + start_a,
	       b + start_b,
	       (stop_b - start_b) * sizeof(int));
	memcpy(dest + start_a + stop_b - start_b,
	       a + stop_a,
	       (len_a - stop_a) * sizeof(int));
}

static void
ls2_crossover(void *a, void *b, void **c, void **d)
{
	ls_genome_t *ga = a;
	ls_genome_t *gb = b;
	ls_genome_t *gc = malloc(sizeof(ls_genome_t));
	ls_genome_t *gd = malloc(sizeof(ls_genome_t));

	int cp = rand() % LS_TYPES;

	for (int i = 0; i < LS_TYPES; i++) {
		if (i < cp) {
			gc->luts[i] = ga->luts[i];
			gd->luts[i] = gb->luts[i];
		} else {
			gd->luts[i] = ga->luts[i];
			gc->luts[i] = gb->luts[i];
		}

		gc->rules[i] = ga->rules[i];
		gd->rules[i] = gb->rules[i];
		/*
		gc->rules[i].len = 1;
		gc->rules[i].succ[0] = i;
		gd->rules[i].len = 1;
		gd->rules[i].succ[0] = i;
		*/
	}

	gc->initial = ga->initial;
	gd->initial = gb->initial;

	int start_a, stop_a;
	int start_b, stop_b;
	int nlen_a, nlen_b;

	for (int i = 0; i < LS_TYPES; i++) {
		int r = rand() % LS_TYPES;

		do {
			int ap = rand() % ga->rules[r].len;
			int bp = rand() % gb->rules[r].len;

			find_subtree(ga->rules[r].succ, ga->rules[r].len, ap, &start_a, &stop_a);
			find_subtree(gb->rules[r].succ, gb->rules[r].len, bp, &start_b, &stop_b);

			nlen_a = ga->rules[r].len + (stop_b - start_b) - (stop_a - start_a);
			nlen_b = gb->rules[r].len + (stop_a - start_a) - (stop_b - start_b);
		} while (nlen_a > LS_SUCCS || nlen_b > LS_SUCCS);

		intersperse_subtree(ga->rules[r].succ, start_a, stop_a, ga->rules[r].len,
							gb->rules[r].succ, start_b, stop_b,
							gc->rules[r].succ);
		intersperse_subtree(gb->rules[r].succ, start_b, stop_b, gb->rules[r].len,
							ga->rules[r].succ, start_a, stop_a,
							gd->rules[r].succ);

		gc->rules[r].len = nlen_a;
		gd->rules[r].len = nlen_b;

		if (rand() % 2)
			break;
	}

	*c = gc;
	*d = gd;
}

ea_cfg_t ls2_cfg =
	{ .new = ls_new
	, .copy = ls_copy
	, .mutate = ls2_mutate
	, .crossover = ls2_crossover
	, .distance = ls_distance
	, .calc_fitness = NULL
	, .user_data = NULL
	, .mut_rate = 0.4
	, .cross_rate = 0.0 // .25
	};

#include "fft.h"

static double
fft_fitness1(ca_t *ca, uint8_t initial)
{
	double tot = 0;
	uint32_t inputs[128];
	double *avg_specs[4];

	for (int i = 0; i < 4; i++) {
		avg_specs[i] = malloc(sizeof(double) * (128 - WIN_SIZE - WIN_SIZE));

		for (int j = 0; j < 128; j++) {
			inputs[j] = ((j / (1 << i)) % 2) ? 0xff : 0x00;
		}

		uint32_t *outs = ca_run_array(ca, inputs, 128, NULL, initial);
		double fft_in[WIN_SIZE];
		double *spec;

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(outs[i + WIN_SIZE - 1]);
		}

		for (int j = WIN_SIZE; j < 128 - WIN_SIZE; j++) {
			memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
			fft_in[WIN_SIZE - 1] = count_bits(outs[j + WIN_SIZE - 1]);

			spec = fft_run(fft_in);
			avg_specs[i][j - WIN_SIZE] = spec_avg(spec, WIN_SIZE / 2 + 1);
		}

		free(outs);
	}

//	tot = dunn_index(avg_specs, 4, 128 - WIN_SIZE - WIN_SIZE);
	tot = linear_regression(avg_specs, 4, 128 - WIN_SIZE - WIN_SIZE);

	for (int i = 0; i < 4; i++) {
		free(avg_specs[i]);
	}

	return tot;
}

static double
fft_fitness2(ca_t *ca, uint8_t initial)
{
	double tot = 0;
	uint32_t inputs[128];
	double *avg_specs[4];

	for (int i = 0; i < 3; i++) {
		avg_specs[i] = malloc(sizeof(double) * (128 - WIN_SIZE - WIN_SIZE));

		for (int j = 0; j < 128; j++) {
			int foo = i < 2 ? i : i + 1;

			inputs[j] = ((j / (1 << foo)) % 2) ? 0xff : 0x00;
		}

		uint32_t *outs = ca_run_array(ca, inputs, 128, NULL, initial);
		double fft_in[WIN_SIZE];
		double *spec;

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(outs[i + WIN_SIZE - 1]);
		}

		for (int j = WIN_SIZE; j < 128 - WIN_SIZE; j++) {
			memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
			fft_in[WIN_SIZE - 1] = count_bits(outs[j + WIN_SIZE - 1]);

			spec = fft_run(fft_in);
			avg_specs[i][j - WIN_SIZE] = spec_avg(spec, WIN_SIZE / 2 + 1);
		}

		free(outs);
	}

//	tot = dunn_index(avg_specs, 3, 128 - WIN_SIZE - WIN_SIZE);
	tot = linear_regression(avg_specs, 3, 128 - WIN_SIZE - WIN_SIZE);

	for (int i = 0; i < 3; i++) {
		free(avg_specs[i]);
	}

	return tot;
}

static double
fft_fitness3(ca_t *ca, uint8_t initial)
{
	double tot = 0;
	uint32_t inputs[128];

	for (int i = 0; i < 10; i++) {
		double fft_in[WIN_SIZE];
		double *spec;
		double avg_spec;

		for (int j = 0; j < WIN_SIZE; j++) {
			inputs[j] = _rand() & 0xff;
			fft_in[j] = count_bits(inputs[j]);
		}
		for (int j = 0; j < WIN_SIZE; j++) {
			inputs[j] = inputs[j % 16];
		}

		spec = fft_run(fft_in);
		avg_spec = spec_avg(spec, WIN_SIZE / 2 + 1);

		uint32_t *outs = ca_run_array(ca, inputs, 128, NULL, initial);

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(outs[i + WIN_SIZE - 1]);
		}

		for (int j = WIN_SIZE; j < 128 - WIN_SIZE; j++) {
			double mean;
			double var;

			memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
			fft_in[WIN_SIZE - 1] = count_bits(outs[j + WIN_SIZE - 1]);

			spec = fft_run(fft_in);
			mean = spec_avg(spec, WIN_SIZE / 2 + 1);
			var = spec_var(spec, WIN_SIZE / 2 + 1) + 1;

			tot += (WIN_SIZE - fabs(mean - avg_spec)) / var;
		}

		free(outs);
	}

	return tot;
}
